
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>

#include "cryptfile.hh"

const char *progname = "crypt";

class PlainFile {
public:
  PlainFile(const char *path, mode_t mode)
    : path_(path), fd_(open(path, mode, 0666)) { check(fd_); }
  int aligned_pread(void *dst, size_t len, size_t offset) {
    return check(::pread(fd_, dst, len, offset));
  }
  int aligned_pwrite(const void *src, size_t len, size_t offset) {
    return check(::pwrite(fd_, src, len, offset));
  }

private:
  const char *path_;
  unique_fd fd_;

  int check(int ret) {
    if (ret == -1)
      threrror(path_);
    return ret;
  }
};

template<typename In, typename Out>
void
copy(In &in, Out &out)
{
  char buf[8192];
  int n = sizeof(buf);
  for (size_t offset = 0; n == sizeof(buf); offset += sizeof(buf)) {
    n = in.aligned_pread(buf, sizeof(buf), offset);
    int m = n;
    if (int remainder = m % CryptFile::blocksize) {
      m += CryptFile::blocksize - remainder;
      memset(buf + n, '\0', m - n);
    }
    out.aligned_pwrite(buf, m, offset);
  }
}

[[noreturn]] void
usage()
{
  fprintf(stderr, "usage: %s {-d | -e} infile outfile\n", progname);
  exit(1);
}

int
main(int argc, char **argv)
{
  if (argc > 0) {
    if ((progname = strrchr(progname, '/')))
      ++progname;
    else
      progname = argv[0];
  }

  bool opt_decrypt;
  if (argc != 4)
    usage();
  if (!strcmp(argv[1], "-d"))
    opt_decrypt = true;
  else if (!strcmp(argv[1], "-e"))
    opt_decrypt = false;
  else
    usage();

  Key key("test");
  try {
    if (opt_decrypt) {
      CryptFile in(key, argv[2]);
      PlainFile out(argv[3], O_CREAT|O_WRONLY|O_TRUNC);
      copy(in, out);
    }
    else {
      PlainFile in(argv[2], O_RDONLY);
      CryptFile out(key, argv[3]);
      copy(in, out);
    }
  }
  catch (std::exception &e) {
    fprintf(stderr, "%s: %s\n", progname, e.what());
    exit(1);
  }
}
