
#include <cassert>
#include <cstdio>
#include <new>

#include "vm.hh"

PhysMem pm(256);

struct AuxPTE {
  const VPage vp;
  const PPage pp;
  Prot prot;
  bool accessed = false;
  bool dirty = false;
  itree_entry link;

  AuxPTE(VPage vp0, Prot p);
  ~AuxPTE();
  void protect(Prot p);
  void clear_accessed() { accessed = false; protect(PROT_NONE); }
};

itree<&AuxPTE::vp, &AuxPTE::link> auxpt;

AuxPTE::AuxPTE(VPage vp0, Prot p)
  : vp(vp0), pp(pm.page_alloc())
{
  if (!pp)
    throw std::runtime_error("out of pages");
  protect(p);
  auxpt.insert(this);
}

AuxPTE::~AuxPTE()
{
  VMRegion::unmap(vp);
  pm.page_free(pp);
}

void
AuxPTE::protect(Prot p)
{
  prot = p;
  VMRegion::map(vp, pp, prot);
  if (prot & PROT_READ)
    accessed = true;
  if (prot & PROT_WRITE)
    dirty = true;
}


struct TrackRegion {
  VMRegion vmem;

  TrackRegion(std::size_t nbytes)
    : vmem(nbytes, [this](char *a){ handler(a); }) {}
  ~TrackRegion();

  void handler(char *addr);

  char &operator[](std::ptrdiff_t i) {
    assert(i >= 0 && std::size_t(i) < vmem.nbytes_);
    return vmem.get_base()[i];
  }
};

TrackRegion::~TrackRegion()
{
  AuxPTE *apte = auxpt.lower_bound(vmem.get_base());
  AuxPTE *end = auxpt.upper_bound(vmem.get_base() + vmem.nbytes_);
  while (apte != end) {
    // IMPORTANT: You must call auxpt.next(apte) before deleting apte!
    AuxPTE *to_delete = apte;
    apte = auxpt.next(apte);
    delete to_delete;
  }
}

static std::string
prot_str(Prot p)
{
  return std::string {
    p & PROT_READ ? 'R' : '-',
    p & PROT_WRITE ? 'W' : '-',
    p & PROT_EXEC ? 'X' : '-',
  };
}

void
TrackRegion::handler(char *addr)
{
  VPage vp = addr - std::intptr_t(addr) % page_size;
  AuxPTE *apte = auxpt[vp];
  if (!apte)
    apte = new AuxPTE(vp, PROT_NONE);
  Prot prot = PROT_READ;
  if (apte->accessed || apte->dirty)
    prot |= PROT_WRITE;
  apte->protect(prot);
  printf("Just mapped %p with prot %s\n", vp, prot_str(prot).c_str());
}

void
dump_pt()
{
  for (AuxPTE &apte : auxpt)
    printf("page %p prot %s is %s\n", apte.vp, prot_str(apte.prot).c_str(),
	   apte.dirty ? "dirty" : apte.accessed ? "accessed" : "untouched");
}

int
main()
{
  constexpr int npgs = 32;
  TrackRegion tr(npgs * page_size);
  char c;
  for (int i = 0; i < npgs; ++i) {
    switch (i % 3) {
    case 0:
      c = tr[i*page_size];
      break;
    case 1:
      tr[i*page_size] = c;
      break;
    }
  }
  printf("==== dumping auxiliary page table ====\n");
  dump_pt();
}
